/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sv_phys.c

#include	"../quakedef.h"
#include	"../game.h"

/*
=============
SV_RunThink

Runs thinking code if time.  There is some play in the exact time the think
function will be called, because it is called before any movement is done
in a frame.  Not used for pushmove objects, because they must be exact.
Returns false if the entity removed itself.
=============
*/
qboolean SV_RunThink (edict_t *ent)
{
	float	thinktime;

	thinktime = ent->game.nextthink;
	if (thinktime <= 0 || thinktime > sv.time + host_frametime)
		return true;
		
	if (thinktime < sv.time)
		thinktime = sv.time;	// don't let things stay in the past.
								// it is possible to start that way
								// by a trigger with a local time.
	ent->game.nextthink		= 0;
	miniracer.self	= ent;
	miniracer.other = sv.edicts;

	ent->game.think ();

	return !ent->free;
}

/*
==================
SV_Impact

Two entities have touched, so run their touch functions
==================
*/
void SV_Impact (edict_t *e1, edict_t *e2)
{
	edict_t	*old_self;
	edict_t	*old_other;

	return;

	old_self	= miniracer.self;
	old_other	= miniracer.other;
	
	if (e1->game.touch && e1->game.solid != SOLID_NOT)
	{
		miniracer.self	= e1;
		miniracer.other	= e2;
		e1->game.touch ();
	}
	
	if (e2->game.touch && e2->game.solid != SOLID_NOT)
	{
		miniracer.self	= e2;
		miniracer.other	= e1;
		e2->game.touch ();
	}

	miniracer.self	= old_self;
	miniracer.other	= old_other;
}


/*
==================
ClipVelocity

Slide off of the impacting object
returns the blocked flags (1 = floor, 2 = step / wall)
==================
*/
#define	STOP_EPSILON	0.1

int ClipVelocity (vec3_t in, vec3_t normal, vec3_t out, float overbounce)
{
	float	backoff;
	float	change;
	int		i, blocked;
	
	blocked = 0;
	if (normal[2] > 0)
		blocked |= 1;		// floor
	if (!normal[2])
		blocked |= 2;		// step
	
	backoff = DotProduct (in, normal) * overbounce;

	for (i=0 ; i<3 ; i++)
	{
		change = normal[i]*backoff;
		out[i] = in[i] - change;
		if (out[i] > -STOP_EPSILON && out[i] < STOP_EPSILON)
			out[i] = 0;
	}
	
	return blocked;
}


/*
============
SV_FlyMove

The basic solid body movement clip that slides along multiple planes
Returns the clipflags if the velocity was modified (hit something solid)
1 = floor
2 = wall / step
4 = dead stop
If steptrace is not NULL, the trace of any vertical wall hit will be stored
============
*/
#define	MAX_CLIP_PLANES	20
int SV_FlyMove (edict_t *ent, float time, trace_t *steptrace)
{
	int			bumpcount, numbumps;
	vec3_t		dir;
	float		d;
	int			numplanes;
	vec3_t		planes[MAX_CLIP_PLANES];
	vec3_t		primal_velocity, original_velocity, new_velocity;
	int			i, j;
	trace_t		trace;
	vec3_t		end;
	float		time_left;
	int			blocked;
	
	numbumps = 4;
	
	blocked = 0;
	VectorCopy (ent->game.velocity, original_velocity);
	VectorCopy (ent->game.velocity, primal_velocity);
	numplanes = 0;

	time_left = time;

	for (bumpcount=0 ; bumpcount<numbumps ; bumpcount++)
	{
		if (!ent->game.velocity[0] && !ent->game.velocity[1] && !ent->game.velocity[2])
			break;

		end[0] = ent->game.origin[0] + time_left * ent->game.velocity[0];
		end[1] = ent->game.origin[1] + time_left * ent->game.velocity[1];
		end[2] = ent->game.origin[2] + time_left * ent->game.velocity[2];

		MR_memset (&trace, 0, sizeof(trace));

		trace = SV_Move (ent->game.origin, ent->game.mins, ent->game.maxs, end, false, ent);

		if (trace.allsolid)
		{	// entity is trapped in another solid
			VectorCopy (vec3_origin, ent->game.velocity);
			return 3;
		}

		if (trace.fraction > 0)
		{	// actually covered some distance
			VectorCopy (trace.endpos, ent->game.origin);
			VectorCopy (ent->game.velocity, original_velocity);
			numplanes = 0;
		}

		if (trace.fraction == 1)
			 break;		// moved the entire distance

		if (!trace.ent)
			Sys_Error ("SV_FlyMove: !trace.ent");

		if (trace.plane.normal[2] > 0.7)
			blocked |= 1;		// floor

		if (!trace.plane.normal[2])
		{
			blocked |= 2;		// step
			if (steptrace)
				*steptrace = trace;	// save for player extrafriction
		}

//
// run the impact function
//

		SV_Impact (ent, trace.ent);
		if (ent->free)
			break;		// removed by the impact function

		time_left -= time_left * trace.fraction;
		
	// cliped to another plane
		if (numplanes >= MAX_CLIP_PLANES)
		{	// this shouldn't really happen
			VectorCopy (vec3_origin, ent->game.velocity);
			return 3;
		}

		VectorCopy (trace.plane.normal, planes[numplanes]);
		numplanes++;

//
// modify original_velocity so it parallels all of the clip planes
//
		for (i=0 ; i<numplanes ; i++)
		{
			ClipVelocity (original_velocity, planes[i], new_velocity, 1);
			for (j=0 ; j<numplanes ; j++)
				if (j != i)
				{
					if (DotProduct (new_velocity, planes[j]) < 0)
						break;	// not ok
				}
			if (j == numplanes)
				break;
		}
		
		if (i != numplanes)
		{	// go along this plane
			VectorCopy (new_velocity, ent->game.velocity);
		}
		else
		{	// go along the crease
			if (numplanes != 2)
			{
//				Con_Printf ("clip velocity, numplanes == %i\n",numplanes);
				VectorCopy (vec3_origin, ent->game.velocity);
				return 7;
			}
			CrossProduct (planes[0], planes[1], dir);
			d = DotProduct (dir, ent->game.velocity);
			VectorScale (dir, d, ent->game.velocity);
		}

//
// if original velocity is against the original velocity, stop dead
// to avoid tiny occilations in sloping corners
//
		if (DotProduct (ent->game.velocity, primal_velocity) <= 0)
		{
			VectorCopy (vec3_origin, ent->game.velocity);
			return blocked;
		}
	}

	return blocked;
}

/*
===============================================================================

CLIENT MOVEMENT

===============================================================================
*/

/*
=====================
SV_WalkMove
======================
*/

void SV_WalkMove (edict_t *ent)
{
	int			clip;
	trace_t		steptrace;

	clip = SV_FlyMove (ent, host_frametime, &steptrace);
}

/*
================
SV_Physics_Client

Player character actions
================
*/
void SV_Physics_Client (edict_t	*ent, int num)
{
	if (ent->game.ClientNo < 0) // Tomaz - To make bots to work
		return;		// unconnected slot

//
// call standard client pre-think
//	
	miniracer.self = ent;

	PlayerPreThink (ent);
	
//
// decide which move function to call
//
	switch (ent->game.movetype)
	{
	case MOVETYPE_NONE:
		if (!SV_RunThink (ent))
			return;
		break;

	case MOVETYPE_WALK:
		if (!SV_RunThink (ent))
			return;

		if (!ent->game.fly)
			ent->game.velocity[2] -= 300 * host_frametime;
	
		SV_WalkMove (ent);
		break;
				
	default:
		Sys_Error ("SV_Physics_client: %i bad movetype %i", (int)num, ent->game.movetype);
	}

//
// call standard player post-think
//		
	SV_LinkEdict (ent, true);

	miniracer.self = ent;

	PlayerPostThink (ent);
}

//============================================================================

/*
===============
SV_Physics_None
===============
*/
void SV_Physics_None (edict_t *ent)
{
	SV_RunThink (ent);
}

//============================================================================

/*
==========
SV_Physics
==========
*/
void SV_Physics (void)
{
	int		i;
	edict_t	*ent;

// let the progs know that a new frame has started
	miniracer.self	= sv.edicts;
	miniracer.other = sv.edicts;

//
// treat each object in turn
//
	ent = sv.edicts;
	ent = NEXT_EDICT(ent);

	for (i=1 ; i<sv.num_edicts ; i++, ent = NEXT_EDICT(ent))
	{
		if (ent->free)
			continue;

		if (i > 0 && i <= svs.maxclients)
			SV_Physics_Client (ent, i);
		else if (ent->game.movetype == MOVETYPE_NONE)
			SV_Physics_None (ent);
		else
			Sys_Error ("SV_Physics: %i bad movetype %i", (int)i, ent->game.movetype);			
	}
	
	sv.time += host_frametime;
}
